package org.processmining.datapetrinets.dsl;

import java.io.InputStream;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;

import org.apache.commons.lang3.text.translate.CharSequenceTranslator;
import org.apache.commons.lang3.text.translate.EntityArrays;
import org.apache.commons.lang3.text.translate.LookupTranslator;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.base.Function;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Ordering;

public final class DPN {

	private static final class DuplicateLabelHandler {

		private final HashMap<PetrinetNode, Integer> labelMap;
		private final Multiset<String> transitionIndex;
		private final Multiset<String> placeIndex;

		private DuplicateLabelHandler() {
			labelMap = new HashMap<>();
			transitionIndex = HashMultiset.create();
			placeIndex = HashMultiset.create();
		}

		public String getLabel(PetrinetNode node) {
			Integer labelIndex = labelMap.get(node);
			if (labelIndex == null) {
				if (node instanceof Transition) {
					labelIndex = transitionIndex.add(node.getLabel(), 1); // count before adding
					labelMap.put(node, labelIndex);
				} else if (node instanceof Place) {
					labelIndex = placeIndex.add(node.getLabel(), 1); // count before adding
					labelMap.put(node, labelIndex);
				}
			}
			if (labelIndex == 0) {
				return quote(node.getLabel());
			} else {
				return quote(node.getLabel()) + "#" + labelIndex;
			}
		}

	}

	private DPN() {
	}

	public static DataPetriNetsWithMarkings fromString(String dpnAsString) throws ParseException {
		return build(new DPNParser(dpnAsString));
	}

	public static DataPetriNetsWithMarkings fromInputStream(InputStream dpnAsStream) throws ParseException {
		return build(new DPNParser(dpnAsStream));
	}

	private static DataPetriNetsWithMarkings build(DPNParser parser) throws ParseException {
		DPNRoot rootNode = parser.parse();
		DPNBuildingVisitor visitor;
		visitor = new DPNBuildingVisitor();
		try {
			visitor.visit(rootNode, null);
		} catch (DPNBuildingError e) {
			if (e.isNestedParseException()) {
				throw new ParseException(e.getToken(), e.getExpectedTokenSequence(), e.getTokenImage());
			} else {
				ParseException parseException = new ParseException("Could not create DPN, error at token "
						+ e.getToken() + "\n Cause: " + e.getMessage());
				parseException.currentToken = e.getToken();
				throw parseException;
			}
		}
		return visitor.getDPN();
	}

	public static String toText(DataPetriNet dpn) {
		return toText(dpn, new Marking(), new Marking());
	}

	public static String toText(DataPetriNetsWithMarkings dpn) {
		return toText(dpn, dpn.getInitialMarking(), (dpn.getFinalMarkings() != null && dpn.getFinalMarkings().length > 0) ? dpn.getFinalMarkings()[0] : null);
	}

	private static final Ordering<PetrinetNode> STANDARD_ORDER = Ordering.natural().nullsFirst()
			.onResultOf(new Function<PetrinetNode, String>() {

				public String apply(PetrinetNode node) {
					return node.getLabel();
				}
			});

	public static String toText(DataPetriNet dpn, Marking initialMarking, Marking finalMarking) {
		DuplicateLabelHandler duplicateHandler = new DuplicateLabelHandler();
		StringBuilder sb = new StringBuilder();
		// name
		sb.append("[NAME]");
		appendNewLine(sb);
		sb.append("\"" + dpn.getLabel() + "\"");
		// place
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[PLACES]");
		for (Place place : STANDARD_ORDER.sortedCopy(dpn.getPlaces())) {
			appendPlace(sb, place, dpn, duplicateHandler);
		}
		// transitions
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[TRANSITIONS]");
		for (Transition transition : STANDARD_ORDER.sortedCopy(dpn.getTransitions())) {
			appendTransition(sb, transition, dpn.getOutEdges(transition), dpn.getInEdges(transition), duplicateHandler);
		}
		// variables
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[VARIABLES]");
		for (DataElement element : STANDARD_ORDER.sortedCopy(dpn.getVariables())) {
			appendNewLine(sb);
			appendVariable(sb, element, duplicateHandler);
		}
		// writes
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[WRITES]");
		for (Transition transition : STANDARD_ORDER.sortedCopy(dpn.getTransitions())) {
			appendWriteOps(sb, transition, dpn.getOutEdges(transition), duplicateHandler);
		}
		// guards
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[GUARDS]");
		for (Transition transition : STANDARD_ORDER.sortedCopy(dpn.getTransitions())) {
			appendGuards(sb, transition, duplicateHandler);
		}
		// initial marking
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[INITIAL]");
		if (initialMarking != null) {
			for (Place place : STANDARD_ORDER.sortedCopy(dpn.getPlaces())) {
				appendMarking(sb, place, initialMarking, duplicateHandler);
			}
		}
		// final marking
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[FINAL]");
		if (finalMarking != null) {
			for (Place place : STANDARD_ORDER.sortedCopy(dpn.getPlaces())) {
				appendMarking(sb, place, finalMarking, duplicateHandler);
			}
		}
		// invisibles
		appendNewLine(sb);
		appendNewLine(sb);
		sb.append("[INVISIBLE]");
		for (Transition t : STANDARD_ORDER.sortedCopy(dpn.getTransitions())) {
			if (t.isInvisible()) {
				appendInvisible(sb, t, duplicateHandler);
			}
		}
		return sb.toString();
	}

	private static void appendInvisible(StringBuilder sb, Transition t, DuplicateLabelHandler duplicateHandler) {
		appendNewLine(sb);
		sb.append(duplicateHandler.getLabel(t));
	}

	private static void appendMarking(StringBuilder sb, Place place, Marking marking,
			DuplicateLabelHandler duplicateHandler) {
		int tokens = marking.occurrences(place);
		if (tokens > 0) {
			appendNewLine(sb);
			sb.append(duplicateHandler.getLabel(place));
			appendSpace(sb);
			sb.append('[');
			sb.append(tokens);
			sb.append(']');
		}
	}

	private static void appendGuards(StringBuilder sb, Transition transition, DuplicateLabelHandler duplicateHandler) {
		if (((PNWDTransition) transition).getGuardExpression() != null) {
			appendNewLine(sb);
			sb.append(duplicateHandler.getLabel(transition));
			appendSpace(sb);
			sb.append("[\"" + escape(((PNWDTransition) transition).getGuardAsString()) + "\"]");
		}
	}

	private static void appendWriteOps(StringBuilder sb, Transition transition,
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> out,
			DuplicateLabelHandler duplicateHandler) {
		boolean hasWrites = hasWrites(out);
		if (hasWrites) {
			appendNewLine(sb);
			sb.append(duplicateHandler.getLabel(transition));
			appendSpace(sb);
			appendWrites(sb, out);
		}
	}

	private static void appendPlace(StringBuilder sb, Place place, DataPetriNet dpn,
			DuplicateLabelHandler duplicateHandler) {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges = dpn.getOutEdges(place);
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : edges) {
			appendNewLine(sb);
			sb.append(duplicateHandler.getLabel(place));
			appendSpace(sb);
			sb.append("->");
			appendSpace(sb);
			sb.append(duplicateHandler.getLabel(edge.getTarget()));
		}
	}

	private static void appendTransition(StringBuilder sb, Transition transition,
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> out,
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> in,
			DuplicateLabelHandler duplicateHandler) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : out) {
			if (!(edge instanceof VariableAccess)) {
				appendNewLine(sb);
				sb.append(duplicateHandler.getLabel(transition));
				appendSpace(sb);
				sb.append("->");
				appendSpace(sb);
				sb.append(duplicateHandler.getLabel(edge.getTarget()));
			}
		}
	}

	private static void appendWrites(StringBuilder sb,
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> writeEdges) {
		sb.append('[');
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : writeEdges) {
			if (edge instanceof VariableAccess) {
				String varName = ((DataElement) edge.getTarget()).getVarName();
				sb.append(quote(varName));
				sb.append(',');
			}
		}
		sb.deleteCharAt(sb.length() - 1);
		sb.append(']');
	}

	private static boolean hasWrites(Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> writeEdges) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : writeEdges) {
			if (edge instanceof VariableAccess) {
				return true;
			}
		}
		return false;
	}

	private static void appendVariable(StringBuilder sb, DataElement element, DuplicateLabelHandler duplicateHandler) {
		String varName = element.getVarName();
		sb.append(quote(varName));
		appendSpace(sb);
		sb.append(convertType(element.getType()));
	}
	
	// Based on StringEscapeUtils from Apache Commons
    private static final CharSequenceTranslator ESCAPE_STRING = 
            new LookupTranslator(
              new String[][] { 
                {"\"", "\\\""},
                {"\\", "\\\\"},
            }).with(
              new LookupTranslator(EntityArrays.JAVA_CTRL_CHARS_ESCAPE())
            );

	private static String escape(String unescaped) {
		return ESCAPE_STRING.translate(unescaped);
	}

	private static String quote(String label) {
		if (label != null && needsQuoting(label)) {
			return "\"" + escape(label) + "\"";
		} else {
			return label;
		}
	}

	private static boolean needsQuoting(String label) {
		if (label.length() > 0) {
			if (Character.isJavaIdentifierStart((label.charAt(0)))) {
				for (int i = 0; i < label.length(); i++) {
					char c = label.charAt(i);
					if (!Character.isJavaIdentifierPart(c)) {
						return true;
					}
				}
				return false;
			}
		}
		return true;
	}

	private static String convertType(Class<?> type) {
		if (type == String.class) {
			return "[LITERAL]";
		} else if (type == Long.class || type == Integer.class) {
			return "[DISCRETE]";
		} else if (type == Boolean.class) {
			return "[BOOLEAN]";
		} else if (type == Double.class || type == Float.class) {
			return "[CONTINUOUS]";
		} else if (type == Date.class) {
			return "[DATE]";
		} else {
			throw new IllegalStateException();
		}
	}

	private static void appendSpace(StringBuilder sb) {
		sb.append(' ');
	}

	private static void appendNewLine(StringBuilder sb) {
		sb.append('\n');
	}

}