package org.processmining.datapetrinets.dsl.editor;

import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;

import org.fife.ui.rtextarea.RTextArea;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Ordering;

public class AutoCompletingRTextArea extends RTextArea {

	private static final long serialVersionUID = -8956607474944135698L;

	private final class DocumentListenerImpl implements DocumentListener {
		
		public void removeUpdate(DocumentEvent e) {
		}

		public void insertUpdate(DocumentEvent e) {
			
			if (e.getLength() != 1) {
				return;
			}

			try {
				int pos = e.getOffset();
				String content = e.getDocument().getText(0, pos + 1);

				int startIndex;
				for (startIndex = pos; startIndex >= 0; startIndex--) {
					if (isStopCharacter(content.charAt(startIndex))) {
						break;
					}
				}
				int charactersInWord = pos - startIndex;
				if (charactersInWord < 2) {
					// Too few chars
					return;
				}

				int startInDocument = startIndex + 1;
				String prefix = e.getDocument().getText(startInDocument, charactersInWord);
				int wordIndex = Collections.binarySearch(getDictionary(), prefix);
				if (wordIndex < 0 && -wordIndex <= getDictionary().size()) {
					String match = getDictionary().get(-wordIndex - 1);
					if (match.startsWith(prefix)) {
						// A completion is found
						String completion = match.substring(pos - startIndex);
						// We cannot modify Document from within notification,
						// so we submit a task that does the change later
						SwingUtilities.invokeLater(new CompletionTask(completion, pos + 1));
					}
				} else {
					// Nothing found
					mode = Mode.INSERT;
				}
			} catch (BadLocationException e1) {
			}
		}

		public void changedUpdate(DocumentEvent e) {
		}
	}

	private static enum Mode {
		INSERT, COMPLETION
	};

	private class CompletionTask implements Runnable {

		String completion;
		int position;

		CompletionTask(String completion, int position) {
			this.completion = completion;
			this.position = position;
		}

		public void run() {
			try {
				getDocument().insertString(position, completion, null);
				setCaretPosition(position + completion.length());
				moveCaretPosition(position);
				getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "commit");
				mode = Mode.COMPLETION;
			} catch (BadLocationException e) {
			}
		}
	}

	private class CommitAction extends AbstractAction {

		private static final long serialVersionUID = 98076970931098392L;

		public void actionPerformed(ActionEvent ev) {
			try {
				if (getSelectedText() != null) {
					if (mode == Mode.COMPLETION) {
						int endPos = getSelectionEnd();
						if (getSelectedText().contains(" ")) {
							int startPos = calcStartPos();
							getDocument().insertString(startPos, "\"", null);							
							endPos++;
							getDocument().insertString(endPos, "\"", null);
							endPos++;
						}					
						getDocument().insertString(endPos, " ", null);
						setCaretPosition(endPos + 1);
						mode = Mode.INSERT;
					} else {
						replaceSelection("");					
					}					
				}
			} catch (BadLocationException e) {
			}
			resetEnterAction();
		}

		private int calcStartPos() throws BadLocationException {
			int selectionStart = getSelectionStart();
			String content = getDocument().getText(0, selectionStart);
			int startIndex;
			for (startIndex = getSelectionStart()-1; startIndex >= 0; startIndex--) {
				if (isStopCharacter(content.charAt(startIndex))) {
					return startIndex+1;
				}
			}
			return 0;
		}

		private void resetEnterAction() {
			getInputMap().put(KeyStroke.getKeyStroke("ENTER"), standardEnterAction);
		}
	}

	private Mode mode = Mode.INSERT;

	private ImmutableList<String> dictionary = ImmutableList.of();
	private ImmutableSet<Character> stopCharacter = ImmutableSet.of(' ', '"', '\n');

	private final Object standardEnterAction;

	public AutoCompletingRTextArea() {
		super();
		//TODO Auto complete is broken
		//getDocument().addDocumentListener(new DocumentListenerImpl());
		//getActionMap().put("commit", new CommitAction());
		standardEnterAction = getInputMap().get(KeyStroke.getKeyStroke("ENTER"));
	}

	public ImmutableList<String> getDictionary() {
		return dictionary;
	}

	/**
	 * @param dictionary a list of words that should be proposed as completion
	 */
	public void setDictionary(List<String> dictionary) {
		this.dictionary = Ordering.natural().immutableSortedCopy(dictionary);
	}

	public ImmutableSet<Character> getStopCharacter() {
		return stopCharacter;
	}

	protected boolean isStopCharacter(char c) {
		return stopCharacter.contains(c);
	}

	public void setStopCharacter(Set<Character> stopCharacter) {
		this.stopCharacter = ImmutableSet.copyOf(stopCharacter);
	}
	
}
