package org.processmining.datapetrinets.dsl.editor;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.Date;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;

import org.fife.ui.rtextarea.RTextArea;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.dsl.DPN;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.plugins.graphviz.visualisation.DotPanel;

import com.google.common.base.Joiner;
import com.google.common.base.Throwables;

final class DPNEditorContextMenu extends JPopupMenu {

	private final class DeleteElement implements ActionListener {
		private final DataPetriNetsWithMarkings net;
		private final PetrinetNode node;
		private final RTextArea textArea;

		private DeleteElement(DataPetriNetsWithMarkings net, PetrinetNode node, RTextArea textArea) {
			this.net = net;
			this.node = node;
			this.textArea = textArea;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = net
						.getOutEdges(node);
				net.getEdges().removeAll(outEdges);
				net.removeNode(node);
				if (node instanceof PNWDTransition) {
					((PNWDTransition) node).setInvisible(false);
				}
				if (node instanceof DataElement) {
					for (Transition t : net.getTransitions()) {
						net.removeWriteOperation(t, (DataElement) node);
						net.removeReadOperation(t, (DataElement) node);
					}
				}
				textArea.setText(DPN.toText(net));
			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private final class ToggleVisibility implements ActionListener {
		private final PetrinetNode node;
		private final RTextArea textArea;
		private final DataPetriNetsWithMarkings net;

		private ToggleVisibility(PetrinetNode node, RTextArea textArea, DataPetriNetsWithMarkings net) {
			this.node = node;
			this.textArea = textArea;
			this.net = net;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				if (node instanceof Transition) {
					((Transition) node).setInvisible(!((Transition) node).isInvisible());
				}
				textArea.setText(DPN.toText(net));
			} finally {
				textArea.endAtomicEdit();
			}
		}
	}

	private final class SetGuard implements ActionListener {
		private final PetrinetNode node;
		private final JTextArea infoMessages;
		private final DataPetriNetsWithMarkings net;
		private final RTextArea textArea;

		private SetGuard(PetrinetNode node, JTextArea infoMessages, DataPetriNetsWithMarkings net, RTextArea textArea) {
			this.node = node;
			this.infoMessages = infoMessages;
			this.net = net;
			this.textArea = textArea;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				PNWDTransition transition = (PNWDTransition) node;
				String existingGuard = transition.hasGuardExpression()
						? transition.getGuardExpression().toCanonicalString() : "";
				String newGuardExpression = JOptionPane.showInputDialog(previewPanel, "Specify guard expression",
						existingGuard);

				if (newGuardExpression != null) {
					try {
						((PNWDTransition) node).setGuard(net, newGuardExpression);
						textArea.setText(DPN.toText(net));
					} catch (NonExistingVariableException e1) {
						infoMessages.setText("Missing variables: " + Joiner.on(',').join(e1.getMissingVariables()));
					} catch (org.processmining.datapetrinets.expression.syntax.ParseException e1) {
						infoMessages.setText("Error parsing expression: " + Throwables.getRootCause(e1).getMessage());
					}
				}

			} finally {
				textArea.endAtomicEdit();
			}
		}
	}

	private final class RenameElement implements ActionListener {
		private final DataPetriNetsWithMarkings net;
		private final RTextArea textArea;
		private final PetrinetNode node;

		private RenameElement(DataPetriNetsWithMarkings net, RTextArea textArea, PetrinetNode node) {
			this.net = net;
			this.textArea = textArea;
			this.node = node;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {

				String newLabel = JOptionPane.showInputDialog(previewPanel, "Specify name", node.getLabel());
				boolean inInitialMarking = false;
				boolean inFinalMarking = false;
				
				if (newLabel != null) {
					
					// Remove node						
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = net
							.getOutEdges(node);
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = net
							.getInEdges(node);
					if (net.getInitialMarking().contains(node)) {
						net.getInitialMarking().remove(node);						
						inInitialMarking = true;
					}	
					if (net.getFinalMarkings() != null && net.getFinalMarkings().length > 0 && net.getFinalMarkings()[0].contains(node)) {
						net.getFinalMarkings()[0].remove(node);						
						inFinalMarking = true;
					}
					net.removeNode(node);

					// Add renamed node
					if (node instanceof Place) {
						Place newPlace = net.addPlace(newLabel);
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : outEdges) {
							net.addArc(newPlace, (Transition) edge.getTarget());
						}
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inEdges) {
							net.addArc((Transition) edge.getSource(), newPlace);
						}
						if (inInitialMarking) {
							net.getInitialMarking().add(newPlace);
						}
						if (inFinalMarking) {
							net.getFinalMarkings()[0].add(newPlace);
						}
					} else if (node instanceof Transition) {
						PNWDTransition newTransition = (PNWDTransition) net.addTransition(newLabel);
						try {
							if (((PNWDTransition) node).hasGuardExpression()) {
								newTransition.setGuard(net, ((PNWDTransition) node).getGuardExpression());
							}
						} catch (NonExistingVariableException e1) {
							throw new RuntimeException(e1);
						}
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : outEdges) {
							if (edge instanceof VariableAccess) {
								net.assignWriteOperation(newTransition, (DataElement) edge.getTarget());
							} else {
								net.addArc(newTransition, (Place) edge.getTarget());
							}
						}
						for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inEdges) {
							if (edge instanceof VariableAccess) {
								net.assignReadOperation(newTransition, (DataElement) edge.getTarget());
							} else {
								net.addArc((Place) edge.getSource(), newTransition);
							}
						}
						newTransition.setInvisible(((Transition) node).isInvisible());
					}
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private final class NewLocalVariableAndWriteOp implements ActionListener {
		private final PetrinetNode node;
		private final RTextArea textArea;
		private final DataPetriNetsWithMarkings net;

		private NewLocalVariableAndWriteOp(PetrinetNode node, RTextArea textArea, DataPetriNetsWithMarkings net) {
			this.node = node;
			this.textArea = textArea;
			this.net = net;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				String varName = JOptionPane.showInputDialog(previewPanel, "Specify variable name");
				Object type = JOptionPane.showInputDialog(previewPanel, "Specify variable type", "Specify variable type",
						JOptionPane.QUESTION_MESSAGE, null, Datatypes.values(), Datatypes.STRING);

				if (varName != null && type != null) {
					DataElement variable = net.addVariable(varName, ((Datatypes) type).getType(), null, null);
					net.assignWriteOperation((Transition) node, variable);
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private class AbstractionAction {

		protected final DataPetriNetsWithMarkings net;
		protected final PetrinetNode node;
		protected final RTextArea textArea;

		private AbstractionAction(DataPetriNetsWithMarkings net, PetrinetNode node, RTextArea textArea) {
			this.net = net;
			this.node = node;
			this.textArea = textArea;
		}

		protected Transition addTransitionIfNeeded(String newLabel) {
			if (existsTransition(net, newLabel)) {
				int result = queryYesNo(previewPanel, "Connect to existing transition?",
						"Do you want to connect to the existing transition '" + newLabel + "'");
				if (result == JOptionPane.YES_OPTION) {
					return getTransition(net, newLabel);
				} else {
					return net.addTransition(newLabel);
				}
			} else {
				return net.addTransition(newLabel);
			}
		}

		protected Place addPlaceIfNeeded(String newLabel) {
			if (existsPlace(net, newLabel)) {
				int result = queryYesNo(previewPanel, "Connect to existing place?",
						"Do you want to connect to the existing place '" + newLabel + "'");
				if (result == JOptionPane.YES_OPTION) {
					return getPlace(net, newLabel);
				} else {
					return net.addPlace(newLabel);
				}
			}
			return net.addPlace(newLabel);
		}

	}

	private final class NewLocalElementSame extends AbstractionAction implements ActionListener {

		private NewLocalElementSame(PetrinetNode node, RTextArea textArea, DataPetriNetsWithMarkings net) {
			super(net, node, textArea);
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				String suggestedLabel = "";
				if (node instanceof Place) {
					suggestedLabel = nextTauPlaceName(net);
				} else if (node instanceof Transition) {
					suggestedLabel = nextTauTransitionName(net);
				}
				String newLabel = JOptionPane.showInputDialog(previewPanel, "Specify name", suggestedLabel);

				if (newLabel != null) {
					// Add renamed node
					if (node instanceof Place) {
						Transition tauTransition = createTauTransition(net);
						net.addArc((Place) node, tauTransition);
						Place newPlace = addPlaceIfNeeded(newLabel);
						net.addArc(tauTransition, newPlace);
					} else if (node instanceof Transition) {
						Place newPlace = createAnonymousPlace(net);
						net.addArc((Transition) node, newPlace);
						Transition newTransition = addTransitionIfNeeded(newLabel);
						net.addArc(newPlace, newTransition);
					}
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private final class NewLocalElementDifferent extends AbstractionAction implements ActionListener {

		private NewLocalElementDifferent(DataPetriNetsWithMarkings net, PetrinetNode node, RTextArea textArea) {
			super(net, node, textArea);
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				String suggestedLabel = "";
				if (node instanceof Place) {
					suggestedLabel = nextTauTransitionName(net);
				} else if (node instanceof Transition) {
					suggestedLabel = nextTauPlaceName(net);
				}
				String newLabel = JOptionPane.showInputDialog(previewPanel, "Specify name", suggestedLabel);

				if (newLabel != null) {
					// Add renamed node
					if (node instanceof Place) {
						Transition transition = addTransitionIfNeeded(newLabel);
						net.addArc((Place) node, transition);
					} else if (node instanceof Transition) {
						Place place = addPlaceIfNeeded(newLabel);
						net.addArc((Transition) node, place);
					}
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}
		}
	}

	private final class NewGlobalVariable implements ActionListener {
		private final RTextArea textArea;
		private final DataPetriNetsWithMarkings net;

		private NewGlobalVariable(RTextArea textArea, DataPetriNetsWithMarkings net) {
			this.textArea = textArea;
			this.net = net;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {

				String varName = JOptionPane.showInputDialog(previewPanel, "Specify variable name");
				Object type = JOptionPane.showInputDialog(previewPanel, "Specify variable type", "Specify variable type",
						JOptionPane.QUESTION_MESSAGE, null, Datatypes.values(), Datatypes.STRING);

				if (varName != null && type != null) {
					net.addVariable(varName, ((Datatypes) type).getType(), null, null);
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private final class NewGlobalPlace implements ActionListener {
		private final RTextArea textArea;
		private final DataPetriNetsWithMarkings net;

		private NewGlobalPlace(RTextArea textArea, DataPetriNetsWithMarkings net) {
			this.textArea = textArea;
			this.net = net;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				String suggestedLabel = nextTauPlaceName(net);
				String newLabel = JOptionPane.showInputDialog(previewPanel, "Specify name", suggestedLabel);

				if (newLabel != null) {
					net.addPlace(newLabel);
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	private final class NewGlobalTransition implements ActionListener {
		private final RTextArea textArea;
		private final DataPetriNetsWithMarkings net;

		private NewGlobalTransition(RTextArea textArea, DataPetriNetsWithMarkings net) {
			this.textArea = textArea;
			this.net = net;
		}

		public void actionPerformed(ActionEvent e) {
			textArea.beginAtomicEdit();
			try {
				String suggestedLabel = nextTauTransitionName(net);
				String newLabel = JOptionPane.showInputDialog(previewPanel, "Specify name", suggestedLabel);

				if (newLabel != null) {
					net.addTransition(newLabel);
					textArea.setText(DPN.toText(net));
				}

			} finally {
				textArea.endAtomicEdit();
			}

		}
	}

	enum Datatypes {

		STRING(String.class), LONG(Long.class), DOUBLE(Double.class), BOOLEAN(Boolean.class), DATE(Date.class);

		private final Class<?> type;

		Datatypes(Class<?> type) {
			this.type = type;

		}

		public Class<?> getType() {
			return type;
		}

	}

	private static final long serialVersionUID = -8797089341134259985L;

	private final DotPanel previewPanel;
	
	private final AtomicInteger automaticPlaceCounter;
	private final AtomicInteger automaticTransitionCounter;

	public DPNEditorContextMenu(final DataPetriNetsWithMarkings net, final PetrinetNode node, final RTextArea textArea,
			final JTextArea infoMessages, DotPanel previewPanel, final AtomicInteger automaticPlaceCounter,
			final AtomicInteger automaticTransitionCounter) {
		super("Options for " + (node != null ? node.getLabel() : "DPN"));
		this.previewPanel = previewPanel;
		this.automaticPlaceCounter = automaticPlaceCounter;
		this.automaticTransitionCounter = automaticTransitionCounter;

		if (node == null) {
			JMenuItem addNewTransitionAction = new JMenuItem("Add Transition");
			addNewTransitionAction.addActionListener(new NewGlobalTransition(textArea, net));
			add(addNewTransitionAction);
			JMenuItem addNewPlaceAction = new JMenuItem("Add Place");
			addNewPlaceAction.addActionListener(new NewGlobalPlace(textArea, net));
			add(addNewPlaceAction);
			JMenuItem addNewVariableAction = new JMenuItem("Add Variable");
			addNewVariableAction.addActionListener(new NewGlobalVariable(textArea, net));
			add(addNewVariableAction);
		} else {

			if (node instanceof Transition || node instanceof Place) {

				JMenuItem addDifferentAction = new JMenuItem("Add " + getOtherTypeLabel(node));
				addDifferentAction.addActionListener(new NewLocalElementDifferent(net, node, textArea));
				add(addDifferentAction);

				JMenuItem addSameAction = new JMenuItem("Add " + getTypeLabel(node));
				addSameAction.addActionListener(new NewLocalElementSame(node, textArea, net));
				add(addSameAction);

				if (node instanceof Transition) {
					JMenuItem addWriteAction = new JMenuItem("Add Write Operation");
					addWriteAction.addActionListener(new NewLocalVariableAndWriteOp(node, textArea, net));
					add(addWriteAction);
				}

				JMenuItem renameAction = new JMenuItem("Rename " + getTypeLabel(node));
				renameAction.addActionListener(new RenameElement(net, textArea, node));
				add(renameAction);
			}

			if (node instanceof Transition) {

				JMenuItem setGuardAction = new JMenuItem("Set Guard");
				setGuardAction.addActionListener(new SetGuard(node, infoMessages, net, textArea));
				add(setGuardAction);

				JMenuItem visibilityAction = new JMenuItem("Toggle visibility");
				visibilityAction.addActionListener(new ToggleVisibility(node, textArea, net));
				add(visibilityAction);
			}

			JMenuItem deleteAction = new JMenuItem("Delete " + getTypeLabel(node));
			deleteAction.addActionListener(new DeleteElement(net, node, textArea));
			add(deleteAction);
		}
	}

	public int queryYesNo(JComponent component, String title, Object message) {
		return JOptionPane.showConfirmDialog(component, message, title, JOptionPane.YES_NO_OPTION);
	}

	public String queryString(JComponent component, String title) {
		return JOptionPane.showInputDialog(component, title);
	}

	private String getTypeLabel(PetrinetNode node) {
		if (node instanceof Transition) {
			return "Transition";
		} else if (node instanceof Place) {
			return "Place";
		} else if (node instanceof DataElement) {
			return "Variable";
		} else {
			return node.getClass().getSimpleName();
		}
	}

	private String getOtherTypeLabel(PetrinetNode node) {
		if (node instanceof Transition) {
			return "Place";
		} else if (node instanceof Place) {
			return "Transition";
		} else {
			throw new IllegalArgumentException("Can only be called for Place and Transition nodes");
		}
	}

	private String nextTauPlaceName(DataPetriNetsWithMarkings net) {
		while (true) {
			int count = automaticPlaceCounter.incrementAndGet();
			String label = "p".concat(String.valueOf(count));
			if (!existsPlace(net, label)) {
				return label;
			}
		}
	}

	private boolean existsPlace(DataPetriNetsWithMarkings net, String label) {
		return getPlace(net, label) != null;
	}

	private Place createAnonymousPlace(DataPetriNetsWithMarkings net) {
		return net.addPlace(nextTauPlaceName(net));
	}

	private Place getPlace(DataPetriNetsWithMarkings net, String label) {
		for (Place p : net.getPlaces()) {
			if (label.equals(p.getLabel())) {
				return p;
			}
		}
		return null;
	}

	private String nextTauTransitionName(DataPetriNetsWithMarkings net) {
		while (true) {
			int count = automaticTransitionCounter.incrementAndGet();
			String label = "t".concat(String.valueOf(count));
			if (!existsTransition(net, label)) {
				return label;
			}
		}
	}

	private Transition createTauTransition(DataPetriNetsWithMarkings net) {
		return net.addTransition(nextTauTransitionName(net));
	}

	private Transition getTransition(DataPetriNetsWithMarkings net, String label) {
		for (Transition t : net.getTransitions()) {
			if (label.equals(t.getLabel())) {
				return t;
			}
		}
		return null;
	}

	private boolean existsTransition(DataPetriNetsWithMarkings net, String label) {
		return getTransition(net, label) != null;
	}

}