package org.processmining.datapetrinets.dsl.editor;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dialog.ModalityType;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingWorker;
import javax.swing.Timer;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Highlighter;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.fife.ui.rtextarea.RTextScrollPane;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.dsl.DPN;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.datapetrinets.utils.TikZUtil;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DPNAsDot;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.GuardDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.PlaceDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.VariableDisplayMode;
import org.processmining.framework.util.ui.widgets.ProMScrollPane;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.DotElementSelectionListener;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Throwables;
import com.google.common.collect.ImmutableList;
import com.kitfox.svg.SVGDiagram;

public final class DPNTextEditor extends JPanel {

	private static final class ModelSaverEditableImpl implements ModelSaver {

		private final Editable<DataPetriNetsWithMarkings> editableDPN;

		private ModelSaverEditableImpl(Editable<DataPetriNetsWithMarkings> editableDPN) {
			this.editableDPN = editableDPN;
		}

		public void save(DataPetriNetsWithMarkings dpnWithMarkings) {
			editableDPN.set(dpnWithMarkings);
		}
	}

	private interface DPNResult {

		SVGDiagram getSVG();

		DPNAsDot getDotResult();

		DataPetriNetsWithMarkings getDPNWithMarkings();

	}

	public interface ModelSaver {
		void save(DataPetriNetsWithMarkings dpnWithMarkings);
	}

	private final static class ElementSelectionListenerColorImpl implements DotElementSelectionListener {

		private final DotPanel dotPanel;

		private String defaultStrokeWidth;
		private String defaultStrokeDashArray;

		private ElementSelectionListenerColorImpl(DotPanel dotPanel, DotElement node) {
			this.dotPanel = dotPanel;
			defaultStrokeWidth = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(dotPanel.getSVG(), node),
					"stroke-width");
			defaultStrokeDashArray = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(dotPanel.getSVG(), node),
					"stroke-dasharray");
		}

		public void colorSelectedElement(final DotPanel dotPanel, DotElement element) {
			defaultStrokeWidth = DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-width", "3");
			defaultStrokeDashArray = DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-dasharray", "5,5");
			dotPanel.repaint();
		}

		public void selected(DotElement element, SVGDiagram image) {
			colorSelectedElement(dotPanel, element);
		}

		public void deselected(DotElement element, SVGDiagram image) {
			DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-width", defaultStrokeWidth);
			DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-dasharray", defaultStrokeDashArray);
			dotPanel.repaint();
		}
	}

	private class DPNParserWorker extends SwingWorker<DPNResult, Void> {

		private final String currentText;
		private final AtomicInteger automaticPlaceCounter;
		private final AtomicInteger automaticTransitionCounter;
		private boolean doSave;
		private boolean firstParse;

		private DPNParserWorker(String currentText, AtomicInteger automaticPlaceCounter,
				AtomicInteger automaticTransitionCounter, boolean doSave, boolean firstParse) {
			this.currentText = currentText;
			this.automaticPlaceCounter = automaticPlaceCounter;
			this.automaticTransitionCounter = automaticTransitionCounter;
			this.doSave = doSave;
			this.firstParse = firstParse;
		}

		@Override
		protected DPNResult doInBackground() throws Exception {
			final DataPetriNetsWithMarkings dpnWithMarkings = DPN.fromString(currentText);
			final DPNAsDot dotResult = DPNGraphvizConverter.convertDPN(dpnWithMarkings,
					dpnWithMarkings.getInitialMarking(), dpnWithMarkings.getFinalMarkings(), GuardDisplayMode.EDGES,
					VariableDisplayMode.NO_LAYOUT, PlaceDisplayMode.DETAILED, GraphDirection.leftRight);
			final SVGDiagram svg = DotPanel.dot2svg(dotResult.getDot());
			return new DPNResult() {

				public DPNAsDot getDotResult() {
					return dotResult;
				}

				public DataPetriNetsWithMarkings getDPNWithMarkings() {
					return dpnWithMarkings;
				}

				public SVGDiagram getSVG() {
					return svg;
				}
			};
		}

		@Override
		protected void done() {
			try {
				if (currentErrorHighlight != null) {
					textArea.getHighlighter().removeHighlight(currentErrorHighlight);
					currentErrorHighlight = null;
				}

				final DPNResult dpnResult = get();
				DPNAsDot dpnAsDot = dpnResult.getDotResult();
				previewPanel.changeDot(dpnAsDot.getDot(), dpnResult.getSVG(), firstParse);
				addListener(dpnResult, dpnAsDot);

				prepareTextArea(dpnResult);

				infoMessages.setForeground(Color.BLACK);
				infoMessages.setText("Updated DPN successfully!");

				if (modelSaver instanceof ModelSaverEditableImpl && doSave) {
					modelSaver.save(dpnResult.getDPNWithMarkings());
				}

			} catch (ExecutionException e) {
				if (e.getCause() instanceof ParseException) {
					ParseException parseEx = (ParseException) e.getCause();
					int beginLine = parseEx.currentToken.next.beginLine - 1;
					int beginColumn = parseEx.currentToken.next.beginColumn - 1;
					int endLine = parseEx.currentToken.next.endLine - 1;
					int endColumn = parseEx.currentToken.next.endColumn - 1;
					try {
						int beginOffset = textArea.getLineStartOffset(beginLine);
						int endOffset = textArea.getLineStartOffset(endLine);
						currentErrorHighlight = textArea.getHighlighter().addHighlight(beginOffset + beginColumn,
								endOffset + endColumn, highlightPainter);
					} catch (BadLocationException e1) {
					}
					infoMessages.setText("Error updating DPN: " + parseEx.getMessage());
					infoMessages.setForeground(Color.RED);
				} else {
					infoMessages.setText("Error updating DPN: " + Throwables.getRootCause(e).getMessage());
					infoMessages.setForeground(Color.RED);
				}
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
			} finally {
				updateFinished();
			}
		}

		private void prepareTextArea(final DPNResult dpnResult) {
			List<String> completions = new ArrayList<>(STANDARD_COMPLETIONS);
			DataPetriNet net = dpnResult.getDPNWithMarkings();
			for (Place p : net.getPlaces()) {
				completions.add(p.getLabel());
			}
			for (Transition t : net.getTransitions()) {
				completions.add(t.getLabel());
			}
			for (DataElement e : net.getVariables()) {
				completions.add(e.getVarName());
			}
			textArea.setDictionary(completions);
		}

		private void addListener(final DPNResult dpnResult, DPNAsDot dpnAsDot) {
			for (Entry<PetrinetNode, DotNode> entry : dpnAsDot.getMapping().entrySet()) {
				final PetrinetNode node = entry.getKey();
				DotNode dotNode = entry.getValue();
				dotNode.addMouseListener(new MouseAdapter() {

					private Popup popup;

					@Override
					public void mousePressed(MouseEvent e) {						
						if (e.isPopupTrigger()) {
							showMenu(e);
						} else {
							if (node instanceof Transition && ((Transition) node).isInvisible()) {
								JLabel label = new JLabel(node.getLabel());
								label.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.LIGHT_GRAY));
								popup = PopupFactory.getSharedInstance().getPopup(e.getComponent(),
										label, e.getXOnScreen(), e.getYOnScreen()-20);
								popup.show();
								Timer timer = new Timer(2000, new ActionListener() {
									
									public void actionPerformed(ActionEvent e) {
										if (popup != null) {
											popup.hide();
											popup = null;
										}
									}
								});
								timer.setRepeats(false);
								timer.start();
							}
						}
					}

					@Override
					public void mouseReleased(MouseEvent e) {
						if (e.isPopupTrigger()) {
							showMenu(e);
						} else {
							if (popup != null) {
								popup.hide();
								popup = null;
							}
						}
					}

					@Override
					public void mouseEntered(MouseEvent e) {
						
					}

					@Override
					public void mouseExited(MouseEvent e) {
					}

					private void showMenu(MouseEvent e) {
						if (!updating) {
							JPopupMenu menu = new DPNEditorContextMenu(dpnResult.getDPNWithMarkings(), node, textArea,
									infoMessages, previewPanel, automaticPlaceCounter, automaticTransitionCounter);
							menu.show(e.getComponent(), e.getX(), e.getY());
						}
					}
				});
				dotNode.addSelectionListener(new ElementSelectionListenerColorImpl(previewPanel, dotNode));
			}
		}

	}	

	private static final long serialVersionUID = -8701706429341608153L;

	private static final int ACTION_DELAY = 1000;

	private final AtomicInteger automaticPlaceCounter = new AtomicInteger();
	private final AtomicInteger automaticTransitionCounter = new AtomicInteger();

	private final AutoCompletingRTextArea textArea;
	private final DotPanel previewPanel;
	private final JTextArea infoMessages;

	private boolean updating = false;
	private final Timer updateTimer;

	private Highlighter.HighlightPainter highlightPainter;
	private Object currentErrorHighlight;

	private final static ImmutableList<String> STANDARD_COMPLETIONS = ImmutableList.of("[NAME]", "[PLACES]",
			"[TRANSITIONS]", "[VARIABLES]", "[WRITES]", "[GUARDS]", "[INITIAL]", "[FINAL]", "[INVISIBLE]", "->");

	private final ModelSaver modelSaver;

	public static DPNTextEditor newInstance(DataPetriNet net, ModelSaver modelSaver) {
		return new DPNTextEditor(DPN.toText(net), modelSaver);
	}

	public static DPNTextEditor newInstance(final Editable<DataPetriNetsWithMarkings> editableDPN) {
		DataPetriNetsWithMarkings dpnWithMarkings = editableDPN.get();
		return new DPNTextEditor(DPN.toText(dpnWithMarkings), new ModelSaverEditableImpl(editableDPN));
	}

	public static DPNTextEditor newInstance(DataPetriNet net, Marking initialMarking, Marking finalMarking,
			ModelSaver modelSaver) {
		return new DPNTextEditor(DPN.toText(net, initialMarking, finalMarking), modelSaver);
	}

	public static DPNTextEditor newInstance(String dpnCode, ModelSaver modelSaver) {
		return new DPNTextEditor(dpnCode, modelSaver);
	}

	private DPNTextEditor(String text, final ModelSaver modelSaver) {
		super();
		this.modelSaver = modelSaver;

		setLayout(new BorderLayout());
		setBackground(Color.WHITE);

		textArea = new AutoCompletingRTextArea();
		textArea.setHighlightCurrentLine(false);
		textArea.setDictionary(STANDARD_COMPLETIONS);

		if (text != null) {
			textArea.setText(text);
			textArea.discardAllEdits();
		}

		highlightPainter = new DefaultHighlighter.DefaultHighlightPainter(Color.RED);
		updateTimer = new Timer(ACTION_DELAY, new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				doUpdate(textArea.getText());
				updateTimer.stop();
			}

		});

		RTextScrollPane scrollPane = new RTextScrollPane(textArea);
		scrollPane.setPreferredSize(new Dimension(Short.MAX_VALUE, 200));

		add(scrollPane, BorderLayout.NORTH);
		textArea.getDocument().addDocumentListener(new DocumentListener() {

			public void removeUpdate(DocumentEvent e) {
				scheduleUpdate();
			}

			public void insertUpdate(DocumentEvent e) {
				if (e.getLength() == e.getDocument().getLength()) {
					try {
						doUpdate(e.getDocument().getText(0, e.getDocument().getLength()));
					} catch (BadLocationException e1) {
						scheduleUpdate();
					}
				} else {
					scheduleUpdate();
				}
			}

			public void changedUpdate(DocumentEvent e) {
				scheduleUpdate();
			}
		});
		Dot dot = new Dot();
		dot.addNode("loading ...");
		previewPanel = new DotPanel(dot);
		previewPanel.setDirection(GraphDirection.leftRight);
		previewPanel.setBackground(Color.WHITE);
		add(previewPanel, BorderLayout.CENTER);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));

		if (!(modelSaver instanceof ModelSaverEditableImpl)) {
			JButton saveButton = SlickerFactory.instance().createButton("Save DPN to ProM Workspace");
			saveButton.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					final String currentText = textArea.getText();
					new SwingWorker<DataPetriNetsWithMarkings, Void>() {

						protected DataPetriNetsWithMarkings doInBackground() throws Exception {
							return DPN.fromString(currentText);
						}

						protected void done() {
							try {
								modelSaver.save(get());
								infoMessages.setText("Sucessfully saved DPN!");
							} catch (InterruptedException | ExecutionException e) {
								infoMessages.setText("Error saving DPN: " + Throwables.getRootCause(e).getMessage());
							}
						}

					}.execute();
				}
			});
			bottomPanel.add(saveButton);
			
			
			JButton saveAsTikzButton = SlickerFactory.instance().createButton("Show TikZ code");
			saveAsTikzButton.addActionListener(new ActionListener() {

				public void actionPerformed(ActionEvent e) {
					final String currentText = textArea.getText();
					new SwingWorker<DataPetriNetsWithMarkings, Void>() {
						
						private static final int DEFAULT_HEIGHT = 800;
						private static final int DEFAULT_WIDTH = 1000;

						protected DataPetriNetsWithMarkings doInBackground() throws Exception {
							return DPN.fromString(currentText);
						}

						protected void done() {
							try {
								DataPetriNetsWithMarkings dpn = get();
								
								TikZUtil tikZUtil = new TikZUtil();
								String tikzCode = tikZUtil.transformToTikZ(dpn, dpn.getInitialMarking() != null ? dpn.getInitialMarking() : new Marking(),
										dpn.getFinalMarkings() != null ? dpn.getFinalMarkings()[0] : new Marking());
								showCustom(new JTextArea(tikzCode), "TikZ code", ModalityType.MODELESS);
								
								infoMessages.setText("Sucessfully saved DPN!");
							} catch (InterruptedException | ExecutionException e) {
								infoMessages.setText("Error saving DPN: " + Throwables.getRootCause(e).getMessage());
							}
						}
						
						private void showCustom(JComponent component, String dialogTitle, ModalityType modalityType) {
							ProMScrollPane scrollPane = new ProMScrollPane(component);
							scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
							if (component.getPreferredSize().getHeight() > DEFAULT_HEIGHT
									|| component.getPreferredSize().getWidth() > DEFAULT_WIDTH) {
								scrollPane.setPreferredSize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
							}
							scrollPane.setMaximumSize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
							scrollPane.setAlignmentX(JComponent.TOP_ALIGNMENT);
							Object[] options = { "OK" };
							JOptionPane optionPane = new JOptionPane(scrollPane, JOptionPane.PLAIN_MESSAGE, JOptionPane.DEFAULT_OPTION, null,
									options, options[0]);
							JDialog dialog = optionPane.createDialog(dialogTitle);
							dialog.setResizable(true);
							dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
							dialog.setModalityType(modalityType);
							dialog.pack();
							dialog.setVisible(true);
						}

					}.execute();
				}
			});
			bottomPanel.add(saveAsTikzButton);			
		
			
		}

		infoMessages = new JTextArea();
		bottomPanel.add(infoMessages);

		add(bottomPanel, BorderLayout.SOUTH);

		new DPNParserWorker(text, automaticPlaceCounter, automaticTransitionCounter, false, true).execute();
	}

	private void doUpdate(String currentText) {
		updateStarted();
		new DPNParserWorker(currentText, automaticPlaceCounter, automaticTransitionCounter, true, false).execute();
	}

	public void updateStarted() {
		updating = true;
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	}

	public void updateFinished() {
		updating = false;
		setCursor(Cursor.getDefaultCursor());
	}

	protected void scheduleUpdate() {
		updateTimer.restart();
	}

	public InteractionResult getUserChoice(UIPluginContext context, String query) {
		return context.showConfiguration(query, this);
	}

	public String getResult() {
		return textArea.getText();
	}

}