package org.processmining.datapetrinets.dsl.editor;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNet.PetrinetWithMarkings;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.dsl.DPN;
import org.processmining.datapetrinets.dsl.editor.DPNTextEditor.ModelSaver;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.utils.ConnectionManagerHelper;
import org.processmining.plugins.utils.ProvidedObjectHelper;

public final class DPNTextEditorPlugin {

	private static final String TEMPLATE = "[NAME]\r\n" 
			+ "\"my petri net\"\r\n" 
			+ "\r\n" + "[PLACES]\r\n"
			+ "p1 -> A\r\n" + "p2 -> B\r\n" 
			+ "p2 -> C\r\n" + "p2 -> skip\r\n" 
			+ "\r\n" + "[TRANSITIONS]\r\n"
			+ "A -> p2\r\n" 
			+ "B -> p3\r\n" 
			+ "C -> p3\r\n" 
			+ "skip -> p3\r\n" 
			+ "\r\n" + "[INITIAL]\r\n" 
			+ "p1 [1]\r\n"
			+ "\r\n" + "[FINAL]\r\n" 
			+ "p3 [1]\r\n" 
			+ "\r\n" 
			+ "[INVISIBLE]\r\n" 
			+ "skip";

	@Plugin(name = "Create Petri net (Text language based)", parameterLabels = {}, returnLabels = {
			"Petri net" }, returnTypes = { DPNTextEditor.class }, //
			userAccessible = true, level = PluginLevel.Regular, help = "Create a Petri net by using a domain specific language.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static DPNTextEditor createDPN(PluginContext context) {
		return createDPNEditor(context.getParentContext(), TEMPLATE);
	}

	@Plugin(name = "Edit Petri net (Text language based)", parameterLabels = { "Petri net" }, returnLabels = {
			"Petri net" }, returnTypes = { DPNTextEditor.class }, //
			userAccessible = true, level = PluginLevel.Regular, help = "Edits a Petri net by using a domain specific language.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static DPNTextEditor editDPN(PluginContext context, Petrinet net) {
		DataPetriNet dpn = convertToDPN(net, context);
		Marking initialMarking = getInitialMarking(context, dpn);
		Marking finalMarking = getFinalMarking(context, dpn);
		return createDPNEditor(context.getParentContext(), DPN.toText(dpn, initialMarking, finalMarking));
	}

	private static DPNTextEditor createDPNEditor(final PluginContext context, String initialCode) {
		return DPNTextEditor.newInstance(initialCode, new ModelSaver() {

			public void save(DataPetriNetsWithMarkings dpn) {
				String netLabel = dpn.getLabel();
				PetrinetWithMarkings pnMarkings = DataPetriNet.Factory.toPetrinetWithMarkings(dpn);
				Petrinet pn = pnMarkings.getNet();
				ProvidedObjectHelper.publish(context, netLabel, pn, Petrinet.class, false);
				ProvidedObjectHelper.setFavorite(context, pn);
				if (pnMarkings.getInitialMarking() != null) {
					ProvidedObjectHelper.publish(context, "Initial marking for " + netLabel,
							pnMarkings.getInitialMarking(), Marking.class, false);
					context.getConnectionManager()
							.addConnection(new InitialMarkingConnection(pn, pnMarkings.getInitialMarking()));
				}
				if (pnMarkings.getFinalMarkings() != null) {
					ProvidedObjectHelper.publish(context, "Final marking for " + netLabel,
							pnMarkings.getFinalMarkings()[0], Marking.class, false);
					context.getConnectionManager()
							.addConnection(new FinalMarkingConnection(pn, pnMarkings.getFinalMarkings()[0]));
				}
			}
		});
	}

	private static Marking getFinalMarking(PluginContext context, PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking[] finalMarkings = ((DataPetriNetsWithMarkings) net).getFinalMarkings();
			if (finalMarkings != null) {
				return finalMarkings[0];
			} else {
				return new Marking();
			}			
		} else {
			try {
				FinalMarkingConnection finalMarkingConnection = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net);
				Marking finalMarking = finalMarkingConnection.getObjectWithRole(FinalMarkingConnection.MARKING);
				return finalMarking;
			} catch (ConnectionCannotBeObtained e) {
				return new Marking();
			}
		}
	}

	private static Marking getInitialMarking(PluginContext context, PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking initialMarking = ((DataPetriNetsWithMarkings) net).getInitialMarking();
			if (initialMarking != null) {
				return initialMarking;
			} else {
				return new Marking();
			}			
		} else {
			try {
				InitialMarkingConnection initialMarkingConnection = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net);
				Marking initialMarking = initialMarkingConnection.getObjectWithRole(InitialMarkingConnection.MARKING);
				return initialMarking;
			} catch (ConnectionCannotBeObtained e) {
				return new Marking();
			}
		}
	}

	private static DataPetriNet convertToDPN(PetrinetGraph net, PluginContext context) {
		if (net instanceof DataPetriNet) {
			return (DataPetriNet) net;
		} else {
			PetriNetWithDataFactory dpnFactory = new PetriNetWithDataFactory(net, net.getLabel(), true);
			dpnFactory.cloneInitialAndFinalConnection(context);
			return dpnFactory.getRetValue();
		}
	}

}
