package org.processmining.datapetrinets.utils;

import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.utils.ConnectionManagerHelper;

import com.google.common.collect.ImmutableList;

public final class MarkingsHelper {

	private MarkingsHelper() {
	}

	/**
	 * Returns the final markings, if available either as part of the Petrinet,
	 * or in the given context. Otherwise returns one empty marking.
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static Marking[] getFinalMarkingsOrEmpty(PluginContext context, PetrinetGraph net) {
		Marking[] markings = tryGetFinalMarkingFromDPN(net);
		if (context != null && (markings == null || markings[0] == null || markings[0].isEmpty())) {
			try {
				return new Marking[] { ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net)
						.getObjectWithRole(FinalMarkingConnection.MARKING) };
			} catch (ConnectionCannotBeObtained e) {
				return new Marking[] { new Marking() };
			}
		} else {
			return markings;
		}
	}

	private static Marking[] tryGetFinalMarkingFromDPN(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking[] finalMarkings = ((DataPetriNetsWithMarkings) net).getFinalMarkings();
			return finalMarkings != null ? finalMarkings : new Marking[] { new Marking() };
		} else {
			return null;
		}
	}

	/**
	 * Returns the initial marking, if available either as part of the Petrinet,
	 * or in the given context. Otherwise returns the empty marking.
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static Marking getInitialMarkingOrEmpty(PluginContext context, PetrinetGraph net) {
		Marking marking = tryGetInitialMarkingFromDPN(net);
		if (context != null && (marking == null || marking.isEmpty())) {
			try {
				return ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net)
						.getObjectWithRole(InitialMarkingConnection.MARKING);
			} catch (ConnectionCannotBeObtained e) {
				return new Marking();
			}
		} else {
			return marking;
		}
	}

	private static Marking tryGetInitialMarkingFromDPN(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking initialMarking = ((DataPetriNetsWithMarkings) net).getInitialMarking();
			return initialMarking != null ? initialMarking : new Marking();
		} else {
			return null;
		}
	}

	/**
	 * Returns true only if the Petrinet has an non-empty initial marking known
	 * in the context or as part of the Petrinet object itself.
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static boolean hasNonEmptyInitialMarking(PluginContext context, PetrinetGraph net) {
		Marking marking = tryGetInitialMarkingFromDPN(net);
		if (marking == null || marking.isEmpty()) {
			try {
				marking = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net)
						.getObjectWithRole(InitialMarkingConnection.MARKING);
				return !marking.isEmpty();
			} catch (ConnectionCannotBeObtained e) {
				return false;
			}
		} else {
			return true;
		}
	}

	/**
	 * Returns true only if the Petrinet has an non-empty final markings known
	 * in the context or as part of the Petrinet object itself.
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static boolean hasNonEmptyFinalMarking(PluginContext context, PetrinetGraph net) {
		Marking[] markings = tryGetFinalMarkingFromDPN(net);
		if (markings == null || (markings.length > 0 && markings[0] != null && markings[0].isEmpty())) {
			try {
				markings = new Marking[] { ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net)
						.getObjectWithRole(FinalMarkingConnection.MARKING) };
				return markings.length > 0 && markings[0] != null && !markings[0].isEmpty();
			} catch (ConnectionCannotBeObtained e) {
				return false;
			}
		} else {
			return true;
		}
	}

	public static Marking guessFinalMarkingByStructure(PetrinetGraph net) {
		Place finalPlace = null;
		for (Place p : net.getPlaces()) {
			if (net.getOutEdges(p).isEmpty()) {
				if (finalPlace != null) {
					return null;
				}
				finalPlace = p;
			}
		}
		if (finalPlace != null) {
			return new Marking(ImmutableList.of(finalPlace));
		}
		return null;
	}

	public static Marking guessInitialMarkingByStructure(PetrinetGraph net) {
		Place initialPlace = null;
		for (Place p : net.getPlaces()) {
			if (net.getInEdges(p).isEmpty()) {
				if (initialPlace != null) {
					return null;
				}
				initialPlace = p;
			}
		}
		if (initialPlace != null) {
			return new Marking(ImmutableList.of(initialPlace));
		}
		return null;
	}

}
