package org.processmining.models.graphbased.directed.petrinetwithdata.newImpl;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.InhibitorArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.ResetArc;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.AbstractResetInhibitorNet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.utils.ProvidedObjectHelper;

import com.google.common.collect.ImmutableMap;

public class PetriNetWithDataFactory {

	private final PetriNetWithData dpn;

	private final Map<Transition, Transition> transMapping = new HashMap<Transition, Transition>();
	private final Map<Place, Place> placeMapping = new HashMap<Place, Place>();

	private PetrinetGraph originalNet;

	private Marking initialMarking = null;
	private Marking finalMarking = null;

	private Map<String, Integer> labelUsedOccurrence = new HashMap<String, Integer>();

	public PetriNetWithDataFactory(PetrinetGraph originalGraph, String netName) {
		this(originalGraph, new PetriNetWithData(netName), false);
	}

	public PetriNetWithDataFactory(PetrinetGraph originalGraph, String netName,
			boolean avoidHavingTransitionWithSameName) {
		this(originalGraph, new PetriNetWithData(netName), avoidHavingTransitionWithSameName);
	}

	/**
	 * @param originalGraph
	 * @param dpn
	 *            the (blank) DPN instance which will be used during the
	 *            conversion
	 * @param avoidHavingTransitionWithSameName
	 */
	public PetriNetWithDataFactory(PetrinetGraph originalGraph, PetriNetWithData dpn,
			boolean avoidHavingTransitionWithSameName) {
		this.dpn = dpn;
		this.originalNet = originalGraph;
		int trNum = 0;
		for (Transition t : originalGraph.getTransitions()) {
			String label = t.getLabel();
			if (label == null || label.trim().equals(""))
				label = "tr" + (++trNum);
			if (avoidHavingTransitionWithSameName) {
				Integer occurrenceLabel = labelUsedOccurrence.get(label);
				if (occurrenceLabel == null)
					occurrenceLabel = 1;
				else
					label = label + occurrenceLabel;
				labelUsedOccurrence.put(t.getLabel(), occurrenceLabel + 1);
			}
			Transition tNew = dpn.addTransition(label);
			if (t.isInvisible()) {
				tNew.setInvisible(true);
			}
			transMapping.put(t, tNew);
		}
		for (Place p : originalGraph.getPlaces()) {
			Place pNew = dpn.addPlace(p.getLabel());
			placeMapping.put(p, pNew);
		}
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> e : originalGraph.getEdges()) {
			PetrinetNode source = e.getSource();
			PetrinetNode target = e.getTarget();
			if (e instanceof Arc) {
				if (source instanceof Place)
					dpn.addArc(placeMapping.get(source), transMapping.get(target));
				else if (source instanceof Transition)
					dpn.addArc(transMapping.get(source), placeMapping.get(target));
			} else if (e instanceof ResetArc) {
				if (source instanceof Place)
					((AbstractResetInhibitorNet) dpn).addResetArc(placeMapping.get(source), transMapping.get(target));
				else if (source instanceof Transition)
					((AbstractResetInhibitorNet) dpn).addResetArc(placeMapping.get(source), transMapping.get(target));
			} else if (e instanceof InhibitorArc) {
				if (source instanceof Place)
					((AbstractResetInhibitorNet) dpn).addInhibitorArc(placeMapping.get(source),
							transMapping.get(target));
				else if (source instanceof Transition)
					((AbstractResetInhibitorNet) dpn).addInhibitorArc(placeMapping.get(source),
							transMapping.get(target));
			}
		}
	}

	private static Marking clone(Marking aMarking, Map<Place, Place> mapping) {
		Marking retValue = new Marking();
		for (Place p : aMarking) {
			Place place = mapping.get(p);
			if (place != null) {
				retValue.add(place);
			} else {
				throw new IllegalArgumentException("Place "+ p+ " is not part of the original net. Known mapping " + mapping);
			}
		}
		return retValue;
	}

	public Marking[] cloneInitialAndFinalConnection(PluginContext context) {

		// Return cached result if available
		if (initialMarking != null || finalMarking != null) {
			dpn.setInitialMarking(initialMarking);
			dpn.setFinalMarkings(new Marking[] { finalMarking });
			return new Marking[] { initialMarking, finalMarking };
		}

		//Copy the initial connections
		try {
			Collection<InitialMarkingConnection> list = context.getConnectionManager()
					.getConnections(InitialMarkingConnection.class, context, originalNet);
			for (InitialMarkingConnection elem : list) {
				MultiSet<Object> connectionObjects = elem.getObjects();
				for (Object objInConnection : connectionObjects) {
					if (objInConnection instanceof Marking) {
						initialMarking = clone((Marking) objInConnection, getPlaceMapping());
						ProvidedObjectHelper.publish(context, "Cloned initial marking " + dpn.getLabel(),
								initialMarking, Marking.class, false);
						InitialMarkingConnection initialConnection = new InitialMarkingConnection(dpn, initialMarking);
						context.getConnectionManager().addConnection(initialConnection);
					}

				}

			}
		} catch (ConnectionCannotBeObtained e) {
		}

		//Copy the final connections
		try {
			Collection<FinalMarkingConnection> list = context.getConnectionManager()
					.getConnections(FinalMarkingConnection.class, context, originalNet);
			for (FinalMarkingConnection elem : list) {
				MultiSet<Object> connectionObjects = elem.getObjects();
				for (Object objInConnection : connectionObjects) {
					if (objInConnection instanceof Marking) {
						finalMarking = clone((Marking) objInConnection, getPlaceMapping());
						ProvidedObjectHelper.publish(context, "Cloned final marking " + dpn.getLabel(), finalMarking,
								Marking.class, false);
						context.getConnectionManager().addConnection(new FinalMarkingConnection(dpn, finalMarking));
					}
				}

			}
		} catch (ConnectionCannotBeObtained e) {
		}

		// Remember markings for DPNWithMarkings interface
		if (initialMarking != null || finalMarking != null) {
			dpn.setInitialMarking(initialMarking);
			dpn.setFinalMarkings(new Marking[] { finalMarking });
		}

		return new Marking[] { initialMarking, finalMarking };
	}

	/**
	 * @return the new DPN based on the supplied Petrinet. Be aware that the
	 *         place and transition objects changed!
	 */
	public PetriNetWithData getRetValue() {
		return dpn;
	}

	/**
	 * @return mapping from old to new transitions
	 */
	public Map<Transition, Transition> getTransMapping() {
		return ImmutableMap.copyOf(transMapping);
	}

	/**
	 * @return mapping from old to new places
	 */
	public Map<Place, Place> getPlaceMapping() {
		return ImmutableMap.copyOf(placeMapping);
	}

	public Marking convertMarking(Marking marking) {
		return clone(marking, placeMapping);
	}
}
